#include <stdarg.h>

#include "console.h"
#include "r8c-ports.h"
#include "globals.h"
#include "fix.h"
#include "i2c.h"
#include "msec.h"
#include "leds.h"

/* Most of the modules are interrupt driven.  However, the 1wire
   peripherals must be polled.  Thus, we use main() to poll them.  */

volatile byte trigger;
volatile byte p1_bits;

static unsigned char onewirebuf[20];
static unsigned char *onewireval;
extern unsigned char ow_wire_therm_addr[8];

unsigned char climate_data[4];

static void
mwait()
{
  static int last_ms = 0;
  while ((msec - last_ms) < 2);
  last_ms = msec;
}

static void
mspause(int ms)
{
  unsigned int start = msec;
  while (msec - start < ms)
    asm("");
}

static int sending_addr = 0;
#if 1
static int need_eol=0;

static void ow_eol()
{
  if (!need_eol)
    return;
  need_eol = 0;
  cput_c ('\n');
}

static int rst()
{
  int i = ow_reset();
  ow_eol();
  printf("rst%d", i);
  need_eol = 1;
  return i;
}

static int rd1()
{
  int i = ow_read ();
  printf(" r\033[33m%d\033[0m", i);
  need_eol = 1;
  return i;
}

static int rd8()
{
  int i = ow_read_byte ();
  printf(" r\033[32m%b\033[0m", i);
  need_eol = 1;
  return i;
}

static int wr8(int b)
{
  int i;
  i = ow_write_byte (b);
  if (i != b)
    printf(" w\[033[31m%b=%b\033[0m", b, i);
  else if (sending_addr)
    printf(" w\033[35m%b\033[0m", b);
  else
    printf(" w\033[34m%b\033[0m", b);
  need_eol = 1;
  return i;
}

#define ow_reset rst
#define ow_read rd1
#define ow_read_byte rd8
#define ow_write_byte wr8
#else
#define ow_eol()
#endif

void
wait_for_complete()
{
  /* Three in case we get a presence pulse.  */
  while (!ow_read());
  while (!ow_read());
  while (!ow_read());
}

int
s16(unsigned char lsb, unsigned char msb)
{
  int rv;
  rv = lsb + (int)msb * 256;
  return rv;
}

#define RETURN0 {printf(" ret:%d\n", __LINE__);return 0;}
/*#define RETURN0 {led_off(GREEN); led_on(RED); return 0;}*/
/*#define RETURN0 return 0*/

int
onewire(int len, ...)
{
  int i, rv = 1;
  int do_wait = 0;
  va_list a;

  onewireval = onewirebuf;

  if (len < 0)
    {
      len = -len;
      do_wait = 1;
    }

  if (ow_reset())
    RETURN0;

  i = ow_write_byte(0x55);
  if (i != 0x55)
    RETURN0;
  sending_addr = 1;
  for (i=0; i<8; i++)
    {
      if (ow_write_byte(ow_search_addr[i]) != ow_search_addr[i])
	{
	  sending_addr = 0;
	  RETURN0;
	}
    }
  sending_addr = 0;

  va_start (a, len);
  for (i=0; i<len; i++)
    {
      int c = va_arg (a, int);
      if (c == -1)
	{
	  onewireval = onewirebuf + i;
	  break;
	}
      onewirebuf[i] = ow_write_byte(c);
      if (onewirebuf[i] != c)
	rv = 0;
    }
  for (; i<len; i++)
    {
      onewirebuf[i] = ow_read_byte();
    }
  if (do_wait)
    wait_for_complete();
  return rv;
}

int
poll_climate ()
{
  int i = 0, p, retry;
  int page[2][8];

  for (p=0; p<2; p++)
    {
      /* write page 0, Vad or Vdd */
      if (!onewire(3, 0x4e, 0x00, p&1 ? 0x08 : 0x00))
	RETURN0;

      /* save page 0 */
      if (!onewire(-2, 0x48, 0x00))
	RETURN0;

      if (p == 0)
	/* initiate T convert */
	if (!onewire(-1, 0x44))
	  RETURN0;

      /* initiate V convert */
      if (!onewire(-1, 0xb4))
	RETURN0;

      /* recall page 0 */
      if (!onewire(2, 0xb8, 0))
	RETURN0;

      for (retry=0; retry<1; retry++)
	{
	  /* read page 0 */
	  if (!onewire (11, 0xbe, 0, -1))
	    {
	      printf("can't read page 0\n");
	      continue;
	    }

	  i = crc(onewireval, 9);
	  if (i != 0)
	    {
	      //printf("bad crc %x %b\n", onewireval, 8, i);
	      continue;
	    }

	  if ((onewireval[0] & 0x08) != ((p&1) ? 0x08 : 0x00))
	    {
	      //printf("oops: %x\n", onewireval, 8);
	      continue;
	    }

	  break;
	}
      if (retry == 5)
	RETURN0;

      for (i=0; i<8; i++)
	page[p][i] = onewireval[i];
    }
  ow_eol();

  int vad = s16(page[0][3], page[0][4]);
  int vdd = s16(page[1][3], page[1][4]);

  if (vdd > 600 || vdd < 200)
    RETURN0;
  if (vad > vdd - 5)
    RETURN0;

  if (page[0][2] == 0xff && page[0][1] == 0xff)
    RETURN0;

  led_on(GREEN);
  led_off(RED);

  FixType temp = fix_make8 (page[0][2], page[0][1]);

  temp = fix_mul (temp, F(1,0.8)); /* 9/5 */
  temp = fix_add (temp, F(32,0));

  /*double humidity1 = ((((double)vad/(double)vdd) - 0.16) / 0.0062) / (1.093 - 0.0012 * temp);*/
  FixType humidity;
  humidity = fix_div (fix_make (vad, 0), fix_make (vdd, 0));
  humidity = fix_sub (humidity, F(0,0.16));
  humidity = fix_mul (humidity, F(161,0.29032)); /* 1/0.0062 */

  FixType correction = fix_sub (F(1,0.093), fix_mul (F(0,0.0012), temp));
  humidity = fix_div (humidity, correction);

  asm("fclr I");
  climate_data[0] = page[0][2]; /* Th */
  climate_data[1] = page[0][1]; /* Tl */
  climate_data[2] = humidity.i;
  climate_data[3] = humidity.f >> 8;
  status_byte |= 0x80;
  asm("fset I");

#if 1
  printf("Temp %F\tVad %w %d Vdd %w %d %F%%\n",
	 temp,
	 vad, vad, vdd, vdd, humidity);
#endif
#if 0
  {
    static char twiddlebites[] = "\\|/-";
    static int twiddle = 0;
    printf("%c\r", twiddlebites[twiddle++]);
    twiddle &= 3;
  }
#endif

  return 1;
}

climate_tx_start()
{
  i2c_tx_buffer = climate_data;
  i2c_tx_count = 4;
  status_byte &= ~0x80;
}

static const i2c_setup_struct climate_command = {
  0, 0, climate_tx_start, 0, 0
};

main()
{
  int i = 0, j, k;
  unsigned int last_poll_time = 0;

  pd1 = 0x9f; /* LEDS output, P1.7 and P1.4 are output. */
  pd3 = 0xff; /* output */
  pd4 = 0x00; /* input */

  drr = 0x0f; /* drive capacity high */

#if 0
  pd1 = 0x07;
  pd3 = 0x00;
  k = 2;
  p1 = 1;
  while (1)
    {
      for (i=0; i<300; i++)
	for (j=0; j<20; j++)
	  asm("");
      switch (k++)
	{
	case 0: p1 = ~ RED; break;
	case 10: p1 = ~ YELLOW; break;
	case 20: p1 = ~ GREEN; break;
	case 30: k = 0; break;
	default: p1 = ~0; break;
	}
    }
#endif

  p1_bits = 0xff;
  p1 = p1_bits;
  p3 = 0x00;

  i2c_init ();
  console_init ();
  msec_init ();
  serial_init ();
  ow_init ();
  leds_init ();

  led_off(RED|GREEN|YELLOW);

  i2c_register_command (3, &climate_command);

  /* This pause ensures that the host syncronizes on our serial
     console, because resetting may have caused glitches.  */
  mspause(2);
  printf("\n\n\033[34mR8C 1-wire diagnostics\033[0m\n");

#if 0
  while (1)
    {
      p1 = 0xff;
      mspause(1);
      p1 = 0x00;
      mspause(1);
    }
#endif

#if 0
  while (1)
    {
      p1_bits &= 0x6f;
      p1_bits |= (i & 0x90);
      p1 = p1_bits;
      mspause(1);
      j = p1;
      p1_bits &= 0xfc;
      if (j & 0x40)
	p1_bits |= 1;
      if (j & 0x20)
	p1_bits |= 2;
      p1 = p1_bits;
      i ^= 0xff;
      mspause(1);
    }
#endif

#if 0
  while (1)
    {
      mspause(1);
      p1_bits |= 0x80;
      p1 = p1_bits;
      mspause(1);
      p1_bits &= ~0x80;
      p1 = p1_bits;
    }
#endif

#if 0
  while (1)
    {
      int i;
      //ow_reset ();
      //ow_write_byte (0x33); /* read rom */
      ow_write_byte (255);
#if 0
      printf("i = %w %w %w", ow_read_byte(), ow_read_byte(), ow_read_byte());
      printf(" %w %w %w", ow_read_byte(), ow_read_byte(), ow_read_byte());
      printf(" %w %w\n", ow_read_byte(), ow_read_byte());
#endif
      mspause(500);
    }
#endif

  while (1)
    {
      int j = 0;
      printf ("\n\n");
      if (ow_reset ())
	{
	  printf("nothing on the bus\n");
	  goto wait;
	}
      printf("something on the bus, searching\n");

      ow_search_init (0);
      while (ow_search_next ())
	{
	  printf("I got %x error %d\n", ow_search_addr, 8, ow_search_error);
	  if (ow_search_addr[0] == 0x26)
	    {
	      printf("polling climate...\n");
	      poll_climate();
	    }
          if (++j > 10)
	    {
	      printf("too many searches, giving up.\n");
	      break;
	    }
	}
      printf("search done\n");

    wait:
      p1_bits ^= 7;
      p1 = p1_bits;
      mspause(1000);
    }
}
